// A command specified by a rule.

/*
 * Copyright 1997 by John D. Ramsdell
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.neu.ccs.jmk;

/**
 * A command specified by a rule.
 * Each command has an operator which is applied to an array of
 * operands.  Each operand is a string or a command variable.
 * When the command is run, an appropriate value is substituted
 * for each command variable by this class.
 * @version November 1997
 * @author John D. Ramsdell
 */
final class Command
{
  private Operator operator;
  private Operand[] operands;	// A array of strings and command variables

  /**
   * Creates a command from an operator and an array of operands.
   */
  Command(Operator operator, Operand[] operands) {
    this.operator = operator;
    this.operands = operands;
  }

  /**
   * Runs a command.
   * It produces an argument list from the operands and then applies
   * the operator to the arguments.  Each command variable is
   * replaced by its value.
   * @param target the origin for the value of '@' command variable
   * @param first the origin for the value of '&lt;' command variable
   * @param newer the origin for the value of '?' command variable
   * @param match the origin for the value of '%' command variable
   * @param justPrinting if true, print but do not run commands
   * @param out the place to write messages
   */
  void run(Rule target, Rule first, Rule[] newer, String match,
	   boolean justPrinting, java.io.PrintWriter out)
       throws CommandFailedException
  {
    // Compute the number of args
    int n = 0;
    for (int i = 0; i < operands.length; i++) {
      switch (operands[i].getChar()) {
      case '?':
	n += newer.length;
	break;
      case '<':			// Note that first is null for
	if (first != null)	// commands with no prerequisites.
	  n++;
	break;
      case '%':			// Note that match is null for
	if (match != null)	// commands not derived from patterns
	  n++;
	break;
      default:
	n++;
      }
    }

    // Load the args
    String[] args = new String[n];
    n = 0;
    for (int i = 0; i < operands.length; i++) {
      switch (operands[i].getChar()) {
      case '@':
	args[n++] = target.getLocalizedTarget();
	break;
      case '<':
	if (first != null)
	  args[n++] = first.getLocalizedTarget();
	break;
      case '?':
	for (int j = 0; j < newer.length; j++)
	  args[n++] = newer[j].getLocalizedTarget();
	break;
      case '%':
	if (match != null)
	  args[n++] = match;
	break;
      default:
	args[n++] = operands[i].getString();
      }
    }

    // Print the command after command variable substitution.
    out.print(operator.getName());
    for (int i = 0; i < args.length; i++)
      out.print(" " + args[i]);
    out.println();

    // Run the command.
    try {
      if (!justPrinting)
	operator.exec(args, out);
    }
    catch (CommandFailedException ex) {
      throw ex;
    }
    catch (Throwable t) {
      throw new CommandFailedException(t.toString());
    }
  }
}
