// Executes a file operation on file operands.

/*
 * Copyright 1997 by John D. Ramsdell
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.neu.ccs.jmk;

import java.io.*;
import java.util.Vector;

/**
 * This operator is used to implement file commands.
 * To be useful, subclasses must implement
 * <code>getName(String)</code> and <code>exec(File, PrintWriter)</code>.
 * @version November 1997
 * @author John D. Ramsdell
 */
abstract class FileOperator
implements Operator
{
  private final static char wildCard = '*';

  /**
   * The actual method implemented by a particular file operator
   * is given by defining this method.
   * @return true if the file operation completed successfully;
   *         false otherwise
   */
  abstract boolean exec(File arg, PrintWriter out);

  /**
   * Applies the file operator to each argument.
   * If an argument contains the wild card character '*',
   * the argument is expanded, and the file operator is applied
   * to each element in the expansion.
   * @param args operands given to the file operator
   * @param out place to write messages
   * @exception CommandFailedException if operation failed
   */
  public void exec(String[] args, PrintWriter out)
       throws CommandFailedException
  {
    for (int i = 0; i < args.length; i++) {
      String fileName = args[i];
      if (fileName.indexOf(wildCard) < 0) { // If no wild card exec
	if (!exec(new File(fileName), out)) {
	  String msg = getName() + " " + fileName +
	    " operation failed";
	  throw new CommandFailedException(msg);
	}
      }
      else			// else glob and then exec
	exec(glob(fileName), out);
    }
  }

  /**
   * This method expands a localized file name that may contain a wild
   * card to an array of file names without the wild card.
   * The file name may contain no more than one wild card, and
   * all file separators in the file name must preceed any wild card.
   * @exception CommandFailedException if wild card is misused
   */
  static String[] glob(String fileName)
       throws CommandFailedException
  {
    int wildCardIndex = fileName.lastIndexOf(wildCard);
    if (wildCardIndex < 0)	// No wild card
      return new String[] { fileName };
    else {
      int separatorIndex = fileName.lastIndexOf(File.separatorChar);
      String suffix = fileName.substring(wildCardIndex + 1);
      String prefix = fileName.substring(0, wildCardIndex);
      if (separatorIndex > wildCardIndex || prefix.indexOf(wildCard) >= 0) {
	String msg = "Cannot expand " + fileName;
	throw new CommandFailedException(msg);
      }
      String dirName = "";
      File dir = new File(System.getProperty("user.dir"));
      if (separatorIndex >= 0) {
	dirName = fileName.substring(0, separatorIndex + 1);
	dir = new File(dir, dirName);
	prefix = fileName.substring(separatorIndex + 1, wildCardIndex);
      }
      String[] list = dir.list();
      if (list == null)
	return new String[0];
      Vector v = new Vector();
      for (int i = 0; i < list.length; i++) { // Filter list
	if (list[i].startsWith(prefix)
	    && list[i].endsWith(suffix)
	    && list[i].indexOf(wildCard) < 0) // Prevent a cycle
	  v.addElement(dirName + list[i]);
      }
      String[] result = new String[v.size()];
      v.copyInto(result);
      return result;
    }
  }

  /**
   * This method expands an array of localized file names that may
   * contain wild cards to an array of file names without wild cards.
   * Wild cards must conform to the rules given above.
   */
  static String[] glob(String[] args)
       throws CommandFailedException
  {
    for (int i = 0; i < args.length; i++) {
      if (args[i].indexOf(wildCard) >= 0) { // If wild card
	Vector v = new Vector();
	for (int j = 0; j < i; j++) // Add previous strings
	  v.addElement(args[j]);
	addStrings(v, glob(args[i])); // Add this arg
	for (int j = i + 1; j < args.length; j++) // Add remaining args
	  if (args[j].indexOf(wildCard) >= 0)
	    addStrings(v, glob(args[j]));
	  else
	    v.addElement(args[j]);
	String[] result = new String[v.size()];
	v.copyInto(result);
	return result;		// Wild cards found
      }
    }
    return args;		// No wild cards found
  }

  private static void addStrings(Vector v, String[] s) {
    for (int i = 0; i < s.length; i++)
      v.addElement(s[i]);
  }

  /**
   * An entry point for testing exec commands.
   * It simply passes the arguments to the exec operation.
   * @param args operands for the command
   */
  public static void main(String[] args) {
    PrintWriter out = new PrintWriter(System.out, true);
    FileOperator operator = new FileOperator() {
      boolean exec(File arg, PrintWriter out) {
	return arg.delete();
      }
      public String getName() {
	return "delete";
      }
    };

    try {
      operator.exec(args, out);
    }
    catch (Throwable t) {
      System.err.println("Internal error: " + t.getMessage());
      t.printStackTrace();
    }
    out.println("File command " + operator.getName() + " completed");
  }
}
