// The loader class for reading and parsing makefiles

/*
 * Copyright 1998 by John D. Ramsdell
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.neu.ccs.jmk;

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;

/**
 * The loader class for reading and parsing makefiles.
 * This class is rather large because it contains the
 * complete makefile parser.
 * @see edu.neu.ccs.jmk.Make
 * @version March 1998
 * @author John D. Ramsdell
 */
class Loader
{
  private Make make;
  private File file;	// For error messages
  private LineNumberReader in;
  private Hashtable identifierTable;// Identifiers are looked up here.
  private PrintWriter out;
  private boolean verbose;

  Loader(Make make, Reader in) {
    this.make = make;
    out = make.getOut();
    verbose = make.isVerbose();
    file = make.getFile();
    this.in = new LineNumberReader(new BufferedReader(in));
    identifierTable = new Hashtable();
  }

  private Loader(Make make,	// Used for processing an
		 File file,	// include directive
		 Reader in,
		 Hashtable identifierTable) {
    this.make = make;
    out = make.getOut();
    verbose = make.isVerbose();
    this.file = file;
    this.in = new LineNumberReader(new BufferedReader(in));
    this.identifierTable = identifierTable;
  }

  // This flag is used to implement conditionals.
  private boolean conditionTrue = true;

  /*
   * Loads a makefile from a reader.
   * @param in the source of the makefile
   * @exception ParseError if an error occurs during loading
   */
  void load()
       throws ParseError
  {
    try {
      int token = scan();
      if (token == EOF_TOKEN) {
	String msg = "No input found";
	throw new ParseError(file, in.getLineNumber(), msg);
      }
      do {
	decl(token);
	token = scan();
      }
      while (token != EOF_TOKEN);
    }
    finally {			// Tidy up.
      try {
	in.close();
      }
      catch (IOException ex) {
      }
    }
  }

  // A vector of strings is often used while parsing.

  // A declaration
  private void decl(int token)
       throws ParseError
  {
    switch (token) {
    case STRING_TOKEN:
      Vector v = new Vector();
      v.addElement(string);
      rule(v);
      break;
    case IDENTIFIER_TOKEN:	// Could be a rule or an assignment
      Identifier id = identifier;
      int saw = peek();
      if (saw == '=') {
	drop();			// Clear out token
	assign(id);
      }
      else
	rule((Vector)id.getValue().clone());
      break;
    case INCLUDE_TOKEN:
      include();
      break;
    case IF_TOKEN:
      conditional();
      break;
    default:
      String msg = "Expecting an identifier, a string, "
	+ "an include, or a conditional, ";
      throw new ParseError(file, in.getLineNumber(),
			   msg + butGot(token));
    }
  }

  // An assignment
  private void assign(Identifier id)
       throws ParseError
  {
    Vector v = new Vector();
    charTermList(v, ';');
    if (conditionTrue) {
      id.setValue(v);
      if (verbose) {
	out.print(id.getName() + " =");
	showTermList(v);
	out.println(";");
      }
    }
  }

  private void showTermList(Vector v) {
    for (int i = 0; i < v.size(); i++) {
      String term = unscanString((String)v.elementAt(i));
      out.print(" " + term);
    }
  }

  // Read a list of strings into v.
  private void list(Vector v)
       throws ParseError
  {
    for (;;) {
      switch (peek()) {
      case STRING_TOKEN:
	drop();
	v.addElement(string);
	break;
      case IDENTIFIER_TOKEN:
	drop();
	addValue(v, identifier.getValue());
	break;
      case '(':
	drop();
	function(v);
	break;
      default:
	return;
      }
    }
  }

  // read a character terminated list of strings.
  private void charTermList(Vector v, char ch)
       throws ParseError
  {
    list(v);
    int token = scan();
    if (token != ch) {
      String msg = "Expecting a '" + ch + "', ";
      throw new ParseError(file, in.getLineNumber(),
			   msg + butGot(token));
    }
  }

  private static void addValue(Vector output, Vector input) {
    for (int i = 0; i < input.size(); i++)
      output.addElement(input.elementAt(i));
  }

  // A string function call
  // This function checks arguments and then dispatches to methods
  // defined in other classes for each function.
  private void function(Vector v)
       throws ParseError
  {
    int token = scan();
    if (token != IDENTIFIER_TOKEN) {
      String msg = "Expecting a string function, ";
      throw new ParseError(file, in.getLineNumber(),
			   msg + butGot(token));
    }
    String function = identifier.getName();
    Vector args = stringFunctionArgs();
    if (function.equals("subst") || function.equals("patsubst")) {
      if (args.size() != 3) {
	String msg = args.size() + " args for " + function +
	  " but expecting 3";
	throw new ParseError(file, in.getLineNumber(), msg);
      }
      Vector v0 = (Vector)args.elementAt(0);
      if (v0.size() != 1) {
	String msg = "Invalid " + function + " pattern";
	throw new ParseError(file, in.getLineNumber(), msg);
      }
      Vector v1 = (Vector)args.elementAt(1);
      if (v1.size() != 1) {
	String msg = "Invalid " + function + " replacement";
	throw new ParseError(file, in.getLineNumber(), msg);
      }
      String pattern = (String)v0.elementAt(0);
      String replacement = (String)v1.elementAt(0);
      if (function.equals("subst"))
	StringUtils.subst(v, pattern, replacement,
			  (Vector)args.elementAt(2));
      else {
	if (!Matcher.isPatternOkay(pattern)) {
	  String msg = "Invalid " + function +
	    " pattern: " + pattern;
	  throw new ParseError(file, in.getLineNumber(), msg);
	}
	else
	  StringUtils.patsubst(v, pattern, replacement,
			       (Vector)args.elementAt(2));
      }
    }
    else if (function.equals("join") || function.equals("equal")) {
      if (args.size() != 2) {
	String msg = args.size() + " args for " + function +
	  " but expecting 2";
	throw new ParseError(file, in.getLineNumber(), msg);
      }
      if (function.equals("join"))
	StringUtils.join(v, (Vector)args.elementAt(0),
			 (Vector)args.elementAt(1));
      else
	StringUtils.equal(v, (Vector)args.elementAt(0),
			  (Vector)args.elementAt(1));
    }
    else if (function.equals("cat")
	     || function.equals("glob")
	     || function.equals("getprop")) {
      if (args.size() != 1) {
	String msg = args.size() + " args for " + function +
	  " but expecting 1";
	throw new ParseError(file, in.getLineNumber(), msg);
      }
      if (function.equals("cat"))
	StringUtils.cat(v, (Vector)args.elementAt(0));
      else if (function.equals("glob")) {
	try {
	  StringUtils.glob(v, (Vector)args.elementAt(0));
	}
	catch (CommandFailedException ex) {
	  throw new ParseError(file, in.getLineNumber(),
			       ex.getMessage());
	}
      }
      else
	StringUtils.getprop(v, (Vector)args.elementAt(0));
    }
    else {
      String msg = "Unknown string function " + function;
      throw new ParseError(file, in.getLineNumber(), msg);
    }
  }

  // Parse the arguments to a function
  private Vector stringFunctionArgs()
       throws ParseError
  {
    Vector args = new Vector();
    for (;;) {
      Vector arg = new Vector();
      list(arg);
      args.addElement(arg);
      int token = scan();
      switch (token) {
      case ')':			// End of function arguments
	return args;
      case ',':			// Next function argument
	break;
      default:
	String msg = "Expecting a ',' or a ')', ";
	throw new ParseError(file, in.getLineNumber(),
			     msg + butGot(token));
      }
    }
  }

  // A rule
  private void rule(Vector targets)
       throws ParseError
  {
    int lineNumber = in.getLineNumber();
    charTermList(targets, ':');
    String[] prerequisites = prerequisiteList();
    Command[] commands = commandList();

    if (conditionTrue)
      for (int i = 0; i < targets.size(); i++)
	make.addRule((String)targets.elementAt(i),
		     prerequisites,
		     commands,
		     lineNumber);
  }

  // Prerequisites
  private String[] prerequisiteList()
       throws ParseError
  {
    Vector v = new Vector();
    charTermList(v, ';');
    String[] prerequisites = new String[v.size()];
    v.copyInto(prerequisites);
    return prerequisites;
  }

  // Commands
  private Command[] commandList()
       throws ParseError
  {
    int token = peek();
    if (token != '{')
      return new Command[0];
    drop();
    Vector v = new Vector();	// A vector of commands
    for (;;) {
      token = scan();
      switch (token) {
      case '}':			// End of commands
	Command[] cmds = new Command[v.size()];
	v.copyInto(cmds);
	return cmds;
      case IDENTIFIER_TOKEN:	// Next command
	v.addElement(commandItem(identifier.getName()));
	break;
      default:
	String msg = "Expecting an identifier, ";
	throw new ParseError(file, in.getLineNumber(),
			     msg + butGot(token));
      }
    }
  }

  // Various operators available for selection.

  private final static ExecOperator exec = new ExecOperator();

  private final static FileOperator delete = new FileOperator() {
    boolean exec(File arg, PrintWriter out) {
      return arg.delete();
    }
    public String getName() {
      return "delete";
    }
  };

  private final static FileOperator mkdir = new FileOperator() {
    boolean exec(File arg, PrintWriter out) {
      return arg.mkdir();
    }
    public String getName() {
      return "mkdir";
    }
  };

  private final static FileOperator mkdirs = new FileOperator() {
    boolean exec(File arg, PrintWriter out) {
      return arg.mkdirs();
    }
    public String getName() {
      return "mkdirs";
    }
  };

  private final static BinaryFileOperator rename = new BinaryFileOperator() {
    boolean exec(File arg1, File arg2, PrintWriter out) {
      if (arg1.equals(arg2))
	return false;
      return arg1.renameTo(arg2);
    }
    public String getName() {
      return "rename";
    }
  };

  private final static BinaryFileOperator copy = new BinaryFileOperator() {
    boolean exec(File arg1, File arg2, PrintWriter out) {
      if (arg1.equals(arg2))
	return false;
      try {
	int bufferSize = 8192;
	byte[] buffer = new byte[bufferSize];
	FileInputStream src = new FileInputStream(arg1);
	FileOutputStream dst = new FileOutputStream(arg2);
	for (;;) {
	  int n = src.read(buffer);
	  if (n == -1) {
	    src.close();
	    dst.close();
	    return true;
	  }
	  dst.write(buffer, 0, n);
	}
      }
      catch (FileNotFoundException ex) {
	out.println("Cannot open " + ex.getMessage());
	return false;
      }
      catch (IOException ex) {
	out.println(ex.toString());
	return false;
      }
    }
    public String getName() {
      return "copy";
    }
  };

  private final static CreateOperator create = new CreateOperator();

  private final static NoteOperator note = new NoteOperator();

  private final static ClassOperator classOperator = new ClassOperator();

  // A single command
  private Command commandItem(String operatorName)
       throws ParseError
  {
    Operator operator;
    if (operatorName.equals(exec.getName()))
      operator = exec;
    else if (operatorName.equals(delete.getName()))
      operator = delete;
    else if (operatorName.equals(mkdir.getName()))
      operator = mkdir;
    else if (operatorName.equals(mkdirs.getName()))
      operator = mkdirs;
    else if (operatorName.equals(rename.getName()))
      operator = rename;
    else if (operatorName.equals(copy.getName()))
      operator = copy;
    else if (operatorName.equals(create.getName()))
      operator = create;
    else if (operatorName.equals(note.getName()))
      operator = note;
    else if (operatorName.equals(classOperator.getName()))
      operator = classOperator;
    else {
      String msg = "Unrecognized command operator: " + operatorName;
      throw new ParseError(file, in.getLineNumber(), msg);
    }
    return new Command(operator, operandList());
  }

  // The operands of a single command
  private Operand[] operandList()
       throws ParseError
  {
    Vector v = new Vector();	// A vector of strings and operands
    for (;;) {
      list(v);			// Collect strings here
      int token = scan();
      switch (token) {
      case ';':			// End of operand list
	Operand[] operands = new Operand[v.size()]; // Convert to
	for (int i = 0; i < v.size(); i++) { // operand array
	  Object object = v.elementAt(i);
	  if (object instanceof String) {
	    String arg = StringUtils.localizePaths((String)object);
	    operands[i] = new Operand(arg);
	  }
	  else
	    operands[i] = (Operand)object;
	}
	return operands;
      case '@':			// Found a command variable
      case '<':
      case '?':
      case '%':
	v.addElement(new Operand((char)token));
	break;
      default:
	String msg = "Expecting an operand list, ";
	throw new ParseError(file, in.getLineNumber(),
			     msg + butGot(token));
      }
    }
  }

  // Inclusion of the makefiles given by a list of strings.
  private void include()
       throws ParseError
  {
    Vector v = new Vector();
    charTermList(v, ';');
    if (conditionTrue) {
      if (verbose) {
	out.print("include");
	showTermList(v);
	out.println(";");
      }
      String parent = file.getParent();	// Used for relative file names
      for (int i = 0; i < v.size(); i++) {
	String incName = StringUtils.localizePaths((String)v.elementAt(i));
	File incFile = new File(incName);
	String incDir = null;
	if (parent != null && !incFile.isAbsolute()) {
	  incDir = parent;
	  incFile = new File(incDir, incName);
	}
	try {
	  FileReader incIn = new FileReader(incFile);
	  new Loader(make, incFile, incIn, identifierTable).load();
	}
	catch (FileNotFoundException ex) {
	  String msg ="Cannot find include file " + incName;
	  if (incDir != null)
	    msg += " in directory " + incDir;
	  throw new ParseError(file, in.getLineNumber(), msg);
	}
      }
    }
  }

  // Conditional makefile reading.
  // This sets the instance variable conditionTrue.
  private void conditional()
       throws ParseError
  {
    boolean changeSense = conditionTrue;
    boolean elseSeen = false;
    Vector v = new Vector();
    String msg;
    list(v);			// Get list for test
    int token = scan();
    if (token != THEN_TOKEN) {
      msg = "Expecting " + THEN_WORD + ", ";
      throw new ParseError(file, in.getLineNumber(),
			   msg + butGot(token));
    }
    if (verbose && changeSense) {
      out.print(IF_WORD);
      showTermList(v);
      out.println(" " + THEN_WORD);
    }
    if (changeSense)
      conditionTrue = v.size() != 0; // Make the test
    for (;;) {
      token = scan();
      switch (token) {
      case EOF_TOKEN:
	if (elseSeen)
	  msg = "Expecting " + END_WORD + ", ";
	else
	  msg = "Expecting " + END_WORD + " or " + ELSE_WORD + ", ";
	throw new ParseError(file, in.getLineNumber(),
			     msg + butGot(token));
      case END_TOKEN:		// Found conditional closer
	conditionTrue = changeSense;
	if (verbose && changeSense)
	  out.println(END_WORD);
	return;
      case ELSE_TOKEN:		// Found alternative
	if (elseSeen) {
	  msg = "Expecting " + END_WORD + ", ";
	  throw new ParseError(file, in.getLineNumber(),
			       msg + butGot(token));
	}
	elseSeen = true;
	if (verbose && changeSense)
	  out.println(ELSE_WORD);
	if (changeSense)
	  conditionTrue = !conditionTrue;
	break;
      default:
	decl(token);		// Found content
      }
    }
  }

  // For error messages
  private String butGot(int token) {
    switch (token) {
    case EOF_TOKEN:
      return "but found [EOF]";
    case STRING_TOKEN:
      return "but found string " + unscanString(string);
    case IDENTIFIER_TOKEN:
      return "but found identifier " + identifier.getName();
    case INCLUDE_TOKEN:
      return "but found the reserved word " + INCLUDE_WORD;
    case IF_TOKEN:
      return "but found the reserved word " + IF_WORD;
    case THEN_TOKEN:
      return "but found the reserved word " + THEN_WORD;
    case ELSE_TOKEN:
      return "but found the reserved word " + ELSE_WORD;
    case END_TOKEN:
      return "but found the reserved word " + END_WORD;
    default:			// must be a character
      return "but found '" + (char)token + "'";
    }
  }

  private static String unscanString(String s) {
    if (s.indexOf('\\') < 0 && s.indexOf('"') < 0)
      return "\"" + s + "\"";
    // This adds in the escape character.
    StringBuffer sb = new StringBuffer();
    sb.append('"');
    for (int i = 0; i < s.length(); i++) {
      char ch = s.charAt(i);
      switch (ch) {
      case '\\':
      case '"':
	sb.append('\\');
	// fall thru okay
      default:
	sb.append(ch);
      }
    }
    sb.append('"');
    return sb.toString();
  }

  /* * * * * *
   * The Makefile lexical analyzer
   * * * * * */

  // This uses a common implementation method for lexical analyzers
  // written in C like languages.  Tokens are ints, and you can
  // guess rest.

  // Character tokens are the character values.
  private final static int EOF_TOKEN = -1;
  private final static int STRING_TOKEN = -2;
  private final static int IDENTIFIER_TOKEN = -3;
  private final static int INCLUDE_TOKEN = -4;
  private final static int IF_TOKEN = -5;
  private final static int THEN_TOKEN = -6;
  private final static int ELSE_TOKEN = -7;
  private final static int END_TOKEN = -8;

  // Reserved words
  private final static String INCLUDE_WORD = "include";
  private final static String IF_WORD = "if";
  private final static String THEN_WORD = "then";
  private final static String ELSE_WORD = "else";
  private final static String END_WORD = "end";

  // Token values
  private String string;
  private Identifier identifier;

  // This is an identifier

  private static class Identifier
  {
    private String name;
    private Vector value = new Vector(); // A vector of strings

    Identifier(String name) {
      this.name = name;
    }

    String getName() {
      return name;
    }

    Vector getValue() {
      return value;
    }

    void setValue(Vector value) {
      this.value = value;
    }
  }

  private void setIdentifier(String name) {
    identifier = (Identifier)identifierTable.get(name);
    if (identifier == null) {
      identifier = new Identifier(name);
      identifierTable.put(name, identifier);
    }
  }

  private int savedToken;
  private boolean isTokenSaved = false;

  /**
   * Returns a token without removing it from the input stream of tokens.
   * @return first token
   * @exception ParseError if scan throws one
   */
  private int peek()
       throws ParseError
  {
    if (!isTokenSaved) {
      savedToken = scan();
      isTokenSaved = true;
    }
    return savedToken;
  }

  /**
   * Drops a saved token.
   */
  private void drop() {
    isTokenSaved = false;
  }

  // Characters can be pushed back into the input stream of chars.
  // This happens after reading an identifier.
  private boolean isCharPushedBack = false;
  private int pushedBackChar;

  /**
   * Returns a token after removing it from the input stream of tokens.
   * @return first token
   * @exception ParseError if syntax error occurs
   */
  private int scan()
       throws ParseError
  {
    try {
      if (isTokenSaved) {	// Token was saved
	isTokenSaved = false;	// because someone
	return savedToken;	// was peeking.
      }

      int ch;
      if (isCharPushedBack) {	// Char pushed back after
	ch = pushedBackChar;	// reading an identifier.
	isCharPushedBack = false; // Use it instead of reading.
      }
      else
	ch = in.read();

      for (;;) {		// Skip leading spaces
	if (ch == -1)
	  return EOF_TOKEN;
	if (!Character.isWhitespace((char)ch))
	  break;
	ch = in.read();
      }

      switch (ch) {
      case ',':
      case ';':
      case ':':
      case '=':
      case '(':
      case ')':
      case '{':
      case '}':
      case '@':
      case '<':
      case '?':
      case '%':
	return ch;
      case '#':			// Found comment character
	if (in.readLine() == null) // Ignore rest of line
	  return EOF_TOKEN;
	else
	  return scan();	// Try again...
      case '"':			// Found a string
	StringBuffer sb = new StringBuffer();
	for (;;) {
	  ch = in.read();
	  switch (ch) {
	  case -1:
	    throw new ParseError(file, in.getLineNumber(),
				 "EOF in string");
	  case '"':
	    string = sb.toString();
	    return STRING_TOKEN;
	  case '\n':
	    throw new ParseError(file, in.getLineNumber(),
				 "newline in string");

	  case '\\':		// Escaped string character
	    ch = in.read();	// Only " and \ can be escaped.
	    switch (ch) {
	    case '"':
	      break;		// Add to string
	    case '\\':
	      break;
	    case -1:		// Error cases follow
	      throw new ParseError(file, in.getLineNumber(),
				   "EOF in string");
	    case '\n':
	      throw new ParseError(file, in.getLineNumber(),
				   "newline in string");
	    default:
	      throw new ParseError(file, in.getLineNumber(),
				   "Cannot escape " + (char)ch);
	    }
	    // Fall thru okay
	  default:
	    sb.append((char)ch);
	  }
	}
      default:			// It must be an identifier.
	if (!Character.isJavaIdentifierStart((char)ch))
	  throw new ParseError(file, in.getLineNumber(),
			       "Unrecognized character " + (char)ch);
	StringBuffer ib = new StringBuffer();
	ib.append((char)ch);
	for (;;) {
	  ch = in.read();
	  if (Character.isJavaIdentifierPart((char)ch))
	    ib.append((char)ch);
	  else {
	    isCharPushedBack = true;
	    pushedBackChar = ch;
	    String word = ib.toString();
	    if (word.equals(INCLUDE_WORD)) // Identify reserved words
	      return INCLUDE_TOKEN;
	    else if (word.equals(IF_WORD))
	      return IF_TOKEN;
	    else if (word.equals(THEN_WORD))
	      return THEN_TOKEN;
	    else if (word.equals(ELSE_WORD))
	      return ELSE_TOKEN;
	    else if (word.equals(END_WORD))
	      return END_TOKEN;
	    else {
	      setIdentifier(word);
	      return IDENTIFIER_TOKEN;
	    }
	  }
	}
      }
    }
    catch (IOException ex) {
      String msg = "I/O exception: " + ex.getMessage();
      throw new ParseError(file, in.getLineNumber(), msg);
    }
  }
}
