// String processing functions.

/*
 * Copyright 1997 by John D. Ramsdell
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.neu.ccs.jmk;

import java.util.Vector;
import java.io.File;

/**
 * String processing functions.
 * These functions are used by the makefile loader to implement
 * its string processing functions.
 * @see edu.neu.ccs.jmk.Make#load(java.io.LineNumberReader)
 * @version November 1997
 * @author John D. Ramsdell
 */
public class StringUtils
{

  // No instances of this class should be constructed.
  private StringUtils() { }

  private final static char genericSeparatorChar = '/';
  private final static char genericPathSeparatorChar = ':';

  /**
   * Localizes file related strings by translating the
   * the generic separator character '/' to the separator
   * character used on this platform, and the generic
   * path separator ':' to the path separator character
   * used on this platform.
   */
  public static String localizePaths(String paths) {
    paths = paths.replace(genericPathSeparatorChar,
			  File.pathSeparatorChar);
    paths = paths.replace(genericSeparatorChar,
			  File.separatorChar);
    return paths;
  }

  /**
   * Generalizes file related strings by performing the inverse
   * translation given by localizePaths.
   */
  public static String generalizePaths(String paths) {
    paths = paths.replace(File.pathSeparatorChar,
			  genericPathSeparatorChar);
    paths = paths.replace(File.separatorChar,
			  genericSeparatorChar);
    return paths;
  }

  /**
   * The patsubst function matches the pattern with the strings
   * in the input.  A string that does not match is copied unchanged
   * to the output.  A string that does match is replaced by a string
   * generated by replacing wild cards in the replacement with the match.
   * When the replacement does not contain a wild card, the match is
   * concatenated with the replacement.
   * @see edu.neu.ccs.jmk.Matcher
   */
  static void patsubst(Vector output,
		       String pattern,
		       String replacement,
		       Vector input) {
    Matcher matcher = new Matcher(pattern);
    for (int i = 0; i < input.size(); i++) {
      String string = (String)input.elementAt(i);
      String match = matcher.match(string);
      if (match != null) {
	if (Matcher.isPattern(replacement))
	  string = Matcher.subst(match, replacement);
	else
	  string = match + replacement;
      }
      output.addElement(string);
    }
  }

  /**
   * Applies stringSubst to each element of the input.
   */
  static void subst(Vector output,
		    String pattern,
		    String replacement,
		    Vector input) {
    for (int i = 0; i < input.size(); i++) {
      String string = (String)input.elementAt(i);
      output.addElement(stringSubst(pattern, replacement, string));
    }
  }

  /**
   * Substitutes the replacement for every non-overlapping occurrence
   * of the pattern in the string.
   */
  static String stringSubst(String pattern,
			    String replacement,
			    String string) {
    int i = string.indexOf(pattern);
    if (i >= 0) {		// Pattern found.
      StringBuffer sb = new StringBuffer();
      do {
	sb.append(string.substring(0, i));
	sb.append(replacement);
	string = string.substring(i + pattern.length());
	i = string.indexOf(pattern);
      }
      while (i >= 0);
      sb.append(string);
      string = sb.toString();
    }
    return string;
  }

  /**
   * Concatenates the strings in the input into one string which
   * is added to the output.  The null string is added to the output
   * when there are no strings in the input.
   */
  static void cat(Vector output, Vector input) {
    String result;
    if (input.size() > 0) {
      result = (String)input.elementAt(0);
      for (int i = 1; i < input.size(); i++)
	result += (String)input.elementAt(i);
    }
    else
      result = "";
    output.addElement(result);
  }

  /**
   * Expands file name that contain the wild card '*' to a list
   * of file names.  The inputs and outputs use generic separators.
   * @see edu.neu.ccs.jmk.FileOperator#glob(String)
   * @exception CommandFailedException if wild card is misused
   */
  static void glob(Vector output, Vector input)
       throws CommandFailedException
  {
    for (int i = 0; i < input.size(); i++) {
      String string = (String)input.elementAt(i);
      String[] globbed = FileOperator.glob(localizePaths(string));
      for (int j = 0; j < globbed.length; j++)
	output.addElement(generalizePaths(globbed[j]));
    }
  }

  /**
   * Each input is used as a key to retrieve data from the system's
   * properties.  The results use generic separators.
   */
  static void getprop(Vector output, Vector input) {
    for (int i = 0; i < input.size(); i++) {
      String string = (String)input.elementAt(i);
      String prop = System.getProperty(string);
      if (prop != null)
	output.addElement(generalizePaths(prop));
    }
  }

  /**
   * Concatenates each string in prefixes with every string in suffixes.
   */
  static void join(Vector output, Vector prefixes, Vector suffixes) {
    for (int i = 0; i < prefixes.size(); i++) {
      String prefix = (String)prefixes.elementAt(i);
      for (int j = 0; j < suffixes.size(); j++) {
	String suffix = (String)suffixes.elementAt(j);
	output.addElement(prefix + suffix);
      }
    }
  }

  /**
   * Returns true if the two input lists are equal.
   * True is a list containing the string "true".
   * False is the empty list.
   */
  static void equal(Vector output, Vector list1, Vector list2) {
    if (list1.size() != list2.size())
      return;
    for (int i = 0; i < list1.size(); i++) {
      String list1Item = (String)list1.elementAt(i);
      String list2Item = (String)list2.elementAt(i);
      if (!list1Item.equals(list2Item))
	return;
    }
    output.addElement("true");
  }
}
